#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  Copyright (c) 2019.       Mike Herbert
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
import logging
import re
import time

from geofinder import DB, Loc, GeoKeys
from geofinder.GeoKeys import Query, Result, Entry, get_soundex


class GeoDB:
    """
    geoname data database.  Add items, look up items, create tables, indices
    """

    def __init__(self, database):
        self.logger = logging.getLogger(__name__)
        self.start = 0
        self.db = DB.DB(database)
        if self.db.err:
            self.logger.error(f"Error! cannot open database {database}.")
            raise ValueError('Cannot open database')
        else:
            self.create_tables()

        self.db.set_speed_pragmas()
        self.db.set_params(select_str='name, country, admin1_id, admin2_id, lat, lon, f_code, geoid', order_str='', limit_str='LIMIT 300')
        self.geoid_main_dict = {}  # Key is GEOID, Value is DB ID for entry
        self.geoid_admin_dict = {}  # Key is GEOID, Value is DB ID for entry
        self.place_type = ''

    def insert(self, geo_row: (), feat_code: str):
        # We split the data into 2  tables, 1) Admin: ADM0/ADM1/ADM2,  and 2) city data

        if feat_code == 'ADM1' or feat_code == 'ADM0' or feat_code == 'ADM2':
            sql = ''' INSERT OR IGNORE INTO admin(name,country, admin1_id,admin2_id,lat,lon,f_code, geoid, sdx)
                      VALUES(?,?,?,?,?,?,?,?,?) '''
            row_id = self.db.execute(sql, geo_row)
            # Add name to dictionary.  Used by AlternateNames for fast lookup during DB build
            self.geoid_admin_dict[geo_row[Entry.ID]] = row_id
        else:
            sql = ''' INSERT OR IGNORE INTO geodata(name, country, admin1_id, admin2_id, lat, lon, f_code, geoid, sdx)
                      VALUES(?,?,?,?,?,?,?,?,?) '''
            row_id = self.db.execute(sql, geo_row)
            # Add name to dictionary.  Used by AlternateNames for fast lookup during DB build
            self.geoid_main_dict[geo_row[Entry.ID]] = row_id

        return row_id

    def create_tables(self):
        # name, country, admin1_id, admin2_id, lat, lon, f_code, geoid
        sql_geodata_table = """CREATE TABLE IF NOT EXISTS geodata    (
                id           integer primary key autoincrement not null,
                name     text,
                country     text,
                admin1_id     text,
                admin2_id text,
                lat      text,
                lon       text,
                f_code      text,
                geoid      text,
                sdx     text
                                    );"""

        # name, country, admin1_id, admin2_id, lat, lon, f_code, geoid
        sql_admin_table = """CREATE TABLE IF NOT EXISTS admin    (
                id           integer primary key autoincrement not null,
                name     text,
                country     text,
                admin1_id     text,
                admin2_id text,
                lat      text,
                lon       text,
                f_code      text,
                geoid      text,
                sdx     text
                                    );"""

        for tbl in [sql_geodata_table, sql_admin_table]:
            self.db.create_table(tbl)

    def lookup_place(self, place: Loc.Loc) -> []:
        """
        Lookup a place in our geoname.org dictionary and update place with Geo_result with lat, long, District, etc
        The dictionary geo_result entry contains: Lat, Long, districtID (County or State or Province ID)
        There can be multiple entries if a city name isnt unique in a country
        """
        self.start = time.time()
        place.result_type = Result.EXACT_MATCH

        if place.place_type == Loc.PlaceType.ADMIN1:
            self.select_admin1(place)
        elif place.place_type == Loc.PlaceType.ADMIN2:
            self.get_admin1_id(place=place)
            self.select_admin2(place)
        elif place.place_type == Loc.PlaceType.COUNTRY:
            self.select_country(place)
        else:
            self.get_admin1_id(place=place)
            self.get_admin2_id(place=place)
            self.select_city(place)

        if place.result_type == Result.EXACT_MATCH and len(place.prefix) > 0:
            place.result_type = Result.PARTIAL_MATCH

    def select_city(self, place: Loc):
        """
        Search for  entry - try the most exact match first, then less exact matches
        """
        lookup_target = place.target
        if len(lookup_target) == 0:
            return
        pattern = self.create_wildcard(lookup_target)
        sdx = get_soundex(lookup_target)
        self.logger.debug(f'CITY lookup. Targ=[{lookup_target}] adm1 id=[{place.admin1_id}]'
                          f' adm2 id=[{place.admin2_id}] iso=[{place.country_iso}] patt =[{pattern}]')

        query_list = []

        if len(place.country_iso) == 0:
            # No country present - try lookup by name.   No other queries
            query_list.append(Query(where="name = ?",
                                    args=(lookup_target,),
                                    result=Result.PARTIAL_MATCH))
            place.georow_list, place.result_type = self.db.process_query_list(from_tbl='main.geodata', query_list=query_list)
            return

        # Build query list - try each query in order until a match is found
        # Start with the most exact match depending on the data provided.
        if len(place.admin1_name) > 0:
            if len(place.admin2_name) > 0:
                # lookup by name, admin1, admin2
                query_list.append(Query(where="name = ? AND country = ? AND admin1_id = ? AND admin2_id = ?",
                                        args=(lookup_target, place.country_iso, place.admin1_id, place.admin2_id),
                                        result=Result.EXACT_MATCH))

            # lookup by name, admin1
            query_list.append(Query(where="name = ? AND country = ? AND admin1_id = ?",
                                    args=(lookup_target, place.country_iso, place.admin1_id),
                                    result=Result.EXACT_MATCH))

            # lookup by wildcard name, admin1
            query_list.append(Query(where="name LIKE ? AND country = ? AND admin1_id = ?",
                                    args=(pattern, place.country_iso, place.admin1_id),
                                    result=Result.PARTIAL_MATCH))
        elif len(place.admin2_name) == 0:
            # lookup by name
            # admin1 and admin2 werent entered, so a match here is an exact match
            query_list.append(Query(where="name = ? AND country = ?",
                                    args=(lookup_target, place.country_iso),
                                    result=Result.EXACT_MATCH))
        elif len(place.admin2_name) > 0:
            # lookup by name
            # admin1 wasnt entered, so a match here is an exact match
            query_list.append(Query(where="name = ? AND admin2_id=? AND country = ?",
                                    args=(lookup_target, place.admin2_id, place.country_iso),
                                    result=Result.EXACT_MATCH))
        else:
            # lookup by name (partial match since user specified admin)
            query_list.append(Query(where="name = ? AND country = ?",
                                    args=(lookup_target, place.country_iso),
                                    result=Result.PARTIAL_MATCH))

        # lookup by name (partial match since user specified admin)
        query_list.append(Query(where="name LIKE ? AND country = ?",
                                args=(pattern, place.country_iso),
                                result=Result.PARTIAL_MATCH))

        # lookup by Soundex (partial match)
        query_list.append(Query(where="sdx = ? AND admin1_id = ? AND country = ?",
                                args=(sdx, place.admin1_id, place.country_iso),
                                result=Result.PARTIAL_MATCH))

        # Try each query in list until we find a match
        place.georow_list, place.result_type = self.db.process_query_list(from_tbl='main.geodata', query_list=query_list)

    def select_admin2(self, place: Loc):
        """Search for Admin2 entry"""
        lookup_target = place.admin2_name
        if len(lookup_target) == 0:
            return
        place.target = lookup_target
        sdx = get_soundex(lookup_target)

        # Try Admin query until we find a match - each query gets less exact
        query_list = [
            Query(where="name = ? AND country = ? AND admin1_id = ? AND f_code=?",
                  args=(lookup_target, place.country_iso, place.admin1_id, 'ADM2'),
                  result=Result.EXACT_MATCH),
            Query(where="name = ? AND country = ? AND f_code=?",
                  args=(lookup_target, place.country_iso, 'ADM2'),
                  result=Result.PARTIAL_MATCH),
            Query(where="name LIKE ? AND country = ? AND f_code=?",
                  args=(self.create_wildcard(lookup_target), place.country_iso, 'ADM2'),
                  result=Result.PARTIAL_MATCH)
        ]

        place.georow_list, place.result_type = self.db.process_query_list(from_tbl='main.admin', query_list=query_list)

        if len(place.georow_list) == 0:
            # Try city rather than County match.
            save_admin2 = place.admin2_name
            place.city1 = place.admin2_name
            place.admin2_name = ''
            self.logger.debug(f'Try admin2 as city: [{place.target}]')

            self.select_city(place)

            if len(place.georow_list) == 0:
                #  not found.  restore admin
                place.admin2_name = save_admin2
                place.city1 = ''
            else:
                # Found match as a City
                place.place_type = Loc.PlaceType.CITY
                return

    def select_admin1(self, place: Loc):
        """Search for Admin1 entry"""
        lookup_target = place.admin1_name

        pattern = self.create_wildcard(lookup_target)
        if len(lookup_target) == 0:
            return
        sdx = get_soundex(lookup_target)

        # self.logger.debug(f'sel adm1 patt={pattern} iso={place.country_iso}')

        # Try each query until we find a match - each query gets less exact
        query_list = [
            Query(where="name = ? AND country = ? AND f_code = ? ",
                  args=(lookup_target, place.country_iso, 'ADM1'),
                  result=Result.EXACT_MATCH),
            Query(where="name LIKE ? AND country = ?  AND f_code = ?",
                  args=(pattern, place.country_iso, 'ADM1'),
                  result=Result.PARTIAL_MATCH),
            Query(where="sdx = ? AND country = ? AND f_code=?",
                  args=(sdx,  place.country_iso, 'ADM1'),
                  result=Result.PARTIAL_MATCH)
        ]
        place.georow_list, place.result_type = self.db.process_query_list(from_tbl='main.admin', query_list=query_list)

        # if nothing found, Try admin1 as city instead
        if len(place.georow_list) == 0:
            save_admin1 = place.admin1_name
            place.city1 = place.admin1_name
            place.target = place.city1
            place.admin1_name = ''
            self.logger.debug(f'Try admin as city: [{place.target}]')

            self.select_city(place)

            if len(place.georow_list) == 0:
                # still not found.  restore admin1
                place.admin1_name = save_admin1
                place.city1 = ''
            else:
                place.place_type = Loc.PlaceType.CITY

    def select_country(self, place: Loc):
        """Search for Admin1 entry"""
        lookup_target = place.country_iso
        if len(lookup_target) == 0:
            return
        sdx = get_soundex(lookup_target)

        # Try each query until we find a match - each query gets less exact
        query_list = [
            Query(where="country = ? AND f_code = ? ",
                  args=(place.country_iso, 'ADM0'),
                  result=Result.EXACT_MATCH),
            Query(where="sdx = ?  AND f_code=?",
                  args=(sdx, 'ADM0'),
                  result=Result.PARTIAL_MATCH)
        ]

        place.georow_list, place.result_type = self.db.process_query_list(from_tbl='main.admin', query_list=query_list)
        # self.build_result_list(place.georow_list, place.event_year)

    def get_admin1_id(self, place: Loc):
        """Search for Admin1 entry"""
        lookup_target = place.admin1_name
        if len(lookup_target) == 0:
            return

        # Try each query until we find a match - each query gets less exact
        query_list = [
            Query(where="name = ? AND country = ? AND f_code = ? ",
                  args=(lookup_target, place.country_iso, 'ADM1'),
                  result=Result.EXACT_MATCH),
            Query(where="name LIKE ? AND country = ?  AND f_code = ?",
                  args=(self.create_wildcard(lookup_target), place.country_iso, 'ADM1'),
                  result=Result.PARTIAL_MATCH),
            Query(where="name = ?  AND f_code = ?",
                  args=(lookup_target, 'ADM1'),
                  result=Result.PARTIAL_MATCH)]

        row_list, res = self.db.process_query_list(from_tbl='main.admin', query_list=query_list)

        if len(row_list) > 0:
            place.admin1_id = row_list[0][Entry.ADM1]
            if place.country_iso == '':
                place.country_iso = row_list[0][Entry.ISO]

    def get_admin2_id(self, place: Loc):
        """Search for Admin1 entry"""
        lookup_target = place.admin2_name
        if len(lookup_target) == 0:
            return

        # Try each query until we find a match - each query gets less exact
        query_list = []
        if len(place.admin1_id) > 0:
            query_list.append(Query(where="name = ? AND country = ? AND admin1_id=? AND f_code=?",
                                    args=(lookup_target, place.country_iso, place.admin1_id, 'ADM2'),
                                    result=Result.EXACT_MATCH))
            query_list.append(Query(where="name LIKE ? AND country = ? and admin1_id = ? AND f_code=?",
                                    args=(self.create_wildcard(lookup_target), place.country_iso, place.admin1_id, 'ADM2'),
                                    result=Result.PARTIAL_MATCH))
        else:
            query_list.append(Query(where="name = ? AND country = ? AND f_code=?",
                                    args=(lookup_target, place.country_iso, 'ADM2'),
                                    result=Result.EXACT_MATCH))
            query_list.append(Query(where="name LIKE ? AND country = ? AND f_code=?",
                                    args=(self.create_wildcard(lookup_target), place.country_iso, 'ADM2'),
                                    result=Result.PARTIAL_MATCH))

        row_list, res = self.db.process_query_list(from_tbl='main.admin', query_list=query_list)

        # self.logger.debug(f'get adm2 id nm={lookup_target} res={row_list}')

        if len(row_list) > 0:
            row = row_list[0]
            place.admin2_id = row[Entry.ADM2]

    def get_admin1_name(self, place: Loc) -> str:
        """Search for Admin1 entry"""
        lookup_target = place.admin1_id
        if len(lookup_target) == 0:
            return ''

        # Try each query until we find a match - each query gets less exact
        query_list = [
            Query(where="admin1_id = ? AND country = ?  AND f_code = ? ",
                  args=(lookup_target, place.country_iso, 'ADM1'),
                  result=Result.EXACT_MATCH)
        ]
        row_list, res = self.db.process_query_list(from_tbl='main.admin', query_list=query_list)

        if len(row_list) > 0:
            row = row_list[0]
            place.admin1_name = row[Entry.NAME]
            return place.admin1_name
        else:
            return ''

    def get_admin2_name(self, place: Loc) -> str:
        """Search for Admin1 entry"""
        lookup_target = place.admin2_id
        if len(lookup_target) == 0:
            return ''

        # Try each query until we find a match - each query gets less exact
        query_list = [
            Query(where="admin2_id = ? AND country = ? AND admin1_id = ?",
                  args=(lookup_target, place.country_iso, place.admin1_id),
                  result=Result.EXACT_MATCH),
            Query(where="admin2_id = ? AND country = ?",
                  args=(lookup_target, place.country_iso),
                  result=Result.PARTIAL_MATCH)]

        row_list, res = self.db.process_query_list(from_tbl='main.admin', query_list=query_list)

        if len(row_list) > 0:
            row = row_list[0]
            place.admin2_name = row[Entry.NAME]
            # self.logger.debug(f'adm2 nm = {place.admin2_name}')
            return place.admin2_name
        else:
            return ''

    def lookup_geoid(self, place: Loc) -> None:
        """Search for GEOID"""
        query_list = [
            Query(where="geoid = ? ",
                  args=(place.target,),
                  result=Result.EXACT_MATCH)
        ]
        place.georow_list, place.result_type = self.db.process_query_list(from_tbl='main.geodata', query_list=query_list)
        if len(place.georow_list) == 0:
            place.georow_list, place.result_type = self.db.process_query_list(from_tbl='main.admin', query_list=query_list)

    def lookup_main_dbid(self, place: Loc) -> None:
        """Search for DB ID"""
        query_list = [
            Query(where="id = ? ",
                  args=(place.target,),
                  result=Result.EXACT_MATCH)
        ]
        place.georow_list, place.result_type = self.db.process_query_list(from_tbl='main.geodata', query_list=query_list)

    def lookup_admin_dbid(self, place: Loc) -> None:
        """Search for DB ID"""
        query_list = [
            Query(where="id = ? ",
                  args=(place.target,),
                  result=Result.EXACT_MATCH)
        ]
        place.georow_list, place.result_type = self.db.process_query_list(from_tbl='main.admin', query_list=query_list)

    def get_country_name(self, iso: str) -> str:
        """ return country name for specified ISO code """
        if len(iso) == 0:
            return ''

        # Try each query until we find a match - each query gets less exact
        query_list = [
            Query(where="country = ? AND f_code = ? ",
                  args=(iso, 'ADM0'),
                  result=Result.EXACT_MATCH)]

        row_list, res = self.db.process_query_list(from_tbl='main.admin', query_list=query_list)

        if len(row_list) > 0:
            res = row_list[0][Entry.NAME]
        return res

    def get_country_iso(self, place: Loc) -> str:
        """ Return ISO code for specified country"""
        lookup_target = place.country_name
        if len(lookup_target) == 0:
            return ''

        # Try each query until we find a match - each query gets less exact
        query_list = [
            Query(where="name = ? AND f_code = ? ",
                  args=(lookup_target, 'ADM0'),
                  result=Result.EXACT_MATCH),
            Query(where="name LIKE ?  AND f_code = ? ",
                  args=(self.create_wildcard(lookup_target), 'ADM0'),
                  result=Result.PARTIAL_MATCH),
            Query(where="sdx = ?  AND f_code = ? ",
                  args=(GeoKeys.get_soundex (lookup_target), 'ADM0'),
                  result=Result.PARTIAL_MATCH)
        ]

        row_list, result_code = self.db.process_query_list(from_tbl='main.admin', query_list=query_list)

        if len(row_list) > 0:
            res = row_list[0][Entry.ISO]
            if len(row_list) == 1:
                place.country_name = row_list[0][Entry.NAME]
        else:
            res = ''

        return res

    def copy_georow_to_place(self, row, place: Loc):
        # Copy data from DB row into Place
        # self.logger.debug(row)
        place.admin1_id = ''
        place.admin2_id = ''
        place.city1 = ''

        place.country_iso = row[Entry.ISO]
        place.country_name = self.get_country_name(row[Entry.ISO])
        place.lat = row[Entry.LAT]
        place.lon = row[Entry.LON]
        place.feature = row[Entry.FEAT]
        place.geoid = row[Entry.ID]

        if place.feature == 'ADM0':
            self.place_type = Loc.PlaceType.COUNTRY
            pass
        elif place.feature == 'ADM1':
            place.admin1_id = row[Entry.ADM1]
            self.place_type = Loc.PlaceType.ADMIN1
        elif place.feature == 'ADM2':
            place.admin1_id = row[Entry.ADM1]
            place.admin2_id = row[Entry.ADM2]
            self.place_type = Loc.PlaceType.ADMIN2
        else:
            place.admin1_id = row[Entry.ADM1]
            place.admin2_id = row[Entry.ADM2]
            place.city1 = row[Entry.NAME]
            self.place_type = Loc.PlaceType.CITY

        place.admin1_name = self.get_admin1_name(place)
        place.admin2_name = self.get_admin2_name(place)

        if place.admin2_name is None:
            place.admin2_name = ''
        if place.admin1_name is None:
            place.admin1_name = ''
        if place.city1 is None:
            place.city1 = ''

    def clear_geoname_data(self):
        # Delete all the geoname data
        for tbl in ['geodata', 'admin']:
            # noinspection SqlWithoutWhere
            self.db.delete_table(tbl)

    def create_geoid_index(self):
        # Create indices
        self.db.create_index(create_table_sql='CREATE INDEX geoid_idx ON geodata(geoid)')
        self.db.create_index(create_table_sql='CREATE INDEX admgeoid_idx ON admin(geoid)')

    def create_indices(self):
        # Create indices
        self.db.create_index(create_table_sql='CREATE INDEX name_idx ON geodata(name)')
        self.db.create_index(create_table_sql='CREATE INDEX country_idx ON geodata(country)')
        self.db.create_index(create_table_sql='CREATE INDEX admin1_idx ON geodata(admin1_id)')
        self.db.create_index(create_table_sql='CREATE INDEX sdx_idx ON geodata(sdx)')

        # self.db.create_index(create_table_sql='CREATE INDEX geoid_idx ON geodata(geoid)')

        self.db.create_index(create_table_sql='CREATE INDEX adm_name_idx ON admin(name)')
        self.db.create_index(create_table_sql='CREATE INDEX adm_country_idx ON admin(country)')
        self.db.create_index(create_table_sql='CREATE INDEX adm_admin1_idx ON admin(admin1_id)')
        self.db.create_index(create_table_sql='CREATE INDEX adm_sdx_idx ON admin(sdx)')

    @staticmethod
    def make_georow(name: str, iso: str, adm1: str, adm2: str, lat: float, lon: float, feat: str, geoid: str, sdx: str) -> ():
        # Return a dummy geo-row
        res = (name, iso, adm1, adm2, lat, lon, feat, geoid, sdx)
        return res

    def get_row_count(self) -> int:
        return self.db.get_row_count()

    @staticmethod
    def create_wildcard(pattern):
        # Create SQL wildcard pattern (convert * to %).  Add % on end
        if '*' in pattern:
            return re.sub(r"\*", "%", pattern)
        else:
            return pattern + '%'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  Copyright (c) 2019.       Mike Herbert
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA

import logging
from tkinter import ttk

from geofinder import ListboxFrame
from geofinder import Widge as Widge

default = ["ADM1", "ADM2", "ADM3", "ADM4", "ADMF", "CH", "CSTL", "CMTY", "EST ", "HSP",
           "HSTS", "ISL", "MSQE", "MSTY", "MT", "MUS", "PAL", "PPL", "PPLA", "PPLA2", "PPLA3", "PPLA4",
           "PPLC", "PPLG", "PPLH", "PPLL", "PPLQ", "PPLX", "PRK", "PRN", "RUIN", "RLG", "STG", "SQR", "SYG", "VAL"]


class SetupFeatureFrame(ListboxFrame.ListboxFrame):
    """
    SetupFeatureList allows users to add or delete items in the Feature List
    The Feature list is the Feature types that we will load from a Geonames.org file.  For example, we will load
    Villages, Squares, Parks, Cemetaries, etc but not load Refineries, Glaciers, and Coral Reefs.
    For details on Feature Types see http://www.geonames.org/export/codes.html
    SetupFeatureList is derived from ListboxFrame:

    ListboxFrame Displays a scrolling list box that is based on a CachedDictionary.
    ListboxFrame defines the overall Grid Layout
    """

    def __init__(self, frame, title, dir_name, cache_filename):
        self.logger = logging.getLogger(__name__)

        # Add these in addition to the standard widgets we inherit from ListBoxFrame
        self.add_button = ttk.Button(frame, text="add", command=self.add_handler, width=ListboxFrame.BUTTON_WIDTH)
        self.feature_label = Widge.CLabel(frame, text="Enter Geoname Feature below and click on Add button to add", style='Info.TLabel')
        self.feature_entry: Widge.CEntry = Widge.CEntry(frame, text="   ", width=15)  # , style='Info.TLabel')
        super().__init__(frame, title, dir_name, cache_filename)

        # If dictionary is empty, load in defaults
        if len(self.dict) == 0:
            self.logger.error('Feature list is empty. loading defaults')
            self.load_defaults()
            super().add_handler()

    def load_defaults(self):
        for item in default:
            self.dict[item] = ""  # Add item to dict
        super().add_handler()

    def configure_widgets(self, frm):
        # Add the standard widgets in ListBoxFrame
        super().configure_widgets(frm)

        # Add these in addition to the standard widgets in ListBoxFrame
        # todo - remove set_grid_position and just use element.grid()
        Widge.Widge.set_grid_position(self.add_button, "country_button", grd=self.grd)
        Widge.Widge.set_grid_position(self.feature_label, "country_label", grd=self.grd)
        Widge.Widge.set_grid_position(self.feature_entry, "country_entry", grd=self.grd)

    def add_handler(self):
        # Allow user to add an item to list.
        val: str = self.feature_entry.get_text()
        self.dict[val.upper()] = ""  # Add item to dict
        super().add_handler()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  Copyright (c) 2019.       Mike Herbert
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA

import glob
import logging
import os
import tkinter as ttk
import webbrowser
from tkinter import *
from tkinter.ttk import *
from typing import Dict

from geofinder import CachedDictionary, GeoKeys, GFStyle
from geofinder import Widge as Widge


class SetupErrorFrame:
    """
    GUI to display status of GeoCoder install - list missing files
     
    """

    def __init__(self, frame, title, dir_name, cache_filename, error):
        self.logger = logging.getLogger(__name__)
        self.file_error = True
        self.title = title
        self.frame = frame
        self.separator = ":"
        self.dirty_flag = False  # Flag to track if data was modified
        self.error = error

        # Load in list from cache file
        self.directory = dir_name
        self.cache_dir = GeoKeys.get_cache_directory(dir_name)
        self.logger.debug(f'SetupStatusFrame dir={dir_name} sub_dir={self.cache_dir} file={cache_filename}')
        self.cache = CachedDictionary.CachedDictionary(self.cache_dir, cache_filename)
        self.cache.read()
        self.error_dict = {}  # Keep a dictionary of errors

        self.supported_countries_cd = CachedDictionary.CachedDictionary(self.cache_dir, "country_list.pkl")
        self.supported_countries_cd.read()
        self.supported_countries_dct: Dict[str, str] = self.supported_countries_cd.dict

        self.logger.debug(f'country list len={len(self.supported_countries_dct)}')

        self.grd = {"title_label": [0, 0, 5, 5, "W"], "scrollbar": [1, 2, 0, 5, "WNS"], "status": [0, 1, 5, 5, "W"], "add_button": [2, 4, 5, 5, "W"],
                    "listbox": [0, 2, 5, 5, "E"], "unused": [2, 3, 5, 5, "W"], "add_entry": [0, 4, 5, 5, "W"], "load_button": [2, 1, 5, 5, "W"],
                    "geoname_button": [2, 1, 5, 5, "E"], "add_label": [0, 3, 5, 5, "EW"]}
        self.title_label = Widge.CLabel(frame, text=self.title, width=80, style='Info.TLabel')
        self.status = Widge.CLabel(frame, text=" ", width=80, style='Highlight.TLabel')
        self.scrollbar = Scrollbar(frame)
        self.listbox = Listbox(frame, width=80, height=20, bg=GFStyle.LT_GRAY, selectmode=MULTIPLE,
                               yscrollcommand=self.scrollbar.set)
        self.add_button = ttk.Button(frame, text="geonames.org", command=self.web_handler, width=12)


        # Configure buttons and widgets
        self.configure_widgets()

        # Display data
        self.load_handler()

    @staticmethod
    def web_handler():
        """ Bring up browser with help text """
        help_base = "https://download.geonames.org/export/dump/"
        webbrowser.open(help_base)

    def check_configuration(self, error_dict, country_dct):
        country_file_len = 6
        file_list = ['allCountries.txt', 'cities500.txt']
        self.file_error = False

        # Ensure that there are some geoname data files
        path = os.path.join(self.directory, "*.txt")
        self.logger.info(f'Geoname path {path}')
        count = 0
        alt_found = False

        for filepath in glob.glob(path):
            # Ignore the two Admin files
            fname = os.path.basename(filepath)
            if len(fname) == country_file_len or fname in file_list:
                count += 1
            if fname == 'alternateNamesV2.txt':
                alt_found = True

        if not alt_found:
            # Alternate Names file is missing
            # error_dict["Optional alternateNamesV2.txt not found"] = ""
            self.logger.warning('optional alternateNamesV2.txt not found')
            # self.file_error = True

        self.logger.debug(f'geoname file count={count}')
        if count == 0:
            # No data files, add error to error dictionary
            error_dict["No Geoname.org data files found. e.g. AllCountries.txt"] = ""
            self.logger.warning('No Geonames files found')
            self.file_error = True

        if self.file_error:
            self.status.set_text("Download AllCountries.zip from geonames.org and place in {}".format(self.directory))

        # Get country list and validate
        self.logger.debug('load countries')
        # self.supported_countries_cd.read()
        res = self.verify_country_list(country_dct)
        if len(res) > 0:
            error_dict[res] = ''

    def load_handler(self):
        self.check_configuration(self.error_dict, self.supported_countries_dct)

        if len(self.error_dict) > 0:
            # Missing files
            for item in sorted(self.error_dict):
                if len(self.error_dict[item]) > 1:
                    self.listbox.insert(END, "{}:   {}".format(item, self.error_dict[item]))
                else:
                    self.listbox.insert(END, "{}".format(item))
        else:
            self.status.set_text("No configuration errors detected")
            # Load in list and display
            self.listbox.delete(0, END)

    def delete_handler(self):
        # Delete selected items in list
        self.delete_items(self.listbox, self.error_dict)
        self.dirty_flag = True

    def add_handler(self):
        # add  item to list
        self.load_handler()  # Reload listbox with new data
        self.dirty_flag = True

    def delete_items(self, lbox, dct):
        # Delete any items in the list that the user selected
        items = lbox.curselection()
        for line in items:
            tokens = lbox.get(line).split(self.separator)
            print(tokens[0])
            dct.pop(tokens[0], None)

        self.load_handler()  # Reload display

    def is_dirty(self):
        # Tells whether the cache was modified
        return self.dirty_flag

    def configure_widgets(self):
        # Grid config for each widget: {name:[col, row, xpad, ypad, sticky]}
        # Create COLUMN 0 widgets
        self.config_grid(self.title_label, "title_label")
        self.config_grid(self.status, "status")

        # Create COLUMN 1 widgets
        self.config_grid(self.listbox, "listbox")
        self.scrollbar.config(command=self.listbox.yview)
        self.config_grid(self.scrollbar, "scrollbar")
        self.config_grid(self.add_button, 'geoname_button')

    @staticmethod
    def verify_country_list(dct):
        if len(dct) == 0:
            return "No countries enabled. Use Countries Tab to add countries"
        elif len(dct) == 1:
            return f"Only {len(dct)} country enabled.  Review in Countries Tab"
        elif len(dct) < 5:
            return f"{len(dct)} countries enabled.  Review in Countries Tab"
        else:
            return ''

    def write(self):
        # Write out cache file
        self.cache.write()

    def config_grid(self, widget, name):
        if self.grd[name][4] == " ":
            widget.grid(row=self.grd[name][1], column=self.grd[name][0], padx=self.grd[name][2], pady=self.grd[name][3])
        else:
            widget.grid(row=self.grd[name][1], column=self.grd[name][0], padx=self.grd[name][2], pady=self.grd[name][3], sticky=self.grd[name][4])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  Copyright (c) 2019.       Mike Herbert
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA

import logging
import os
import pickle
from pathlib import Path

from geofinder import CachedDictionary, GeoKeys
from geofinder.Widge import Widge


class Config:
    def __init__(self):
        self.logger = logging.getLogger(__name__)
        self.config_cd: CachedDictionary
        self.config_cd = None

        self.directory: str = os.path.join(str(Path.home()), GeoKeys.get_directory_name())
        self.cache_dir = GeoKeys.get_cache_directory(self.directory)

    def get(self, param) -> str:
        res = self.config_cd.dict.get(param)
        if res is None:
            res = ''
        return res

    def set(self, name: str, val: str):
        self.config_cd.dict[name] = val

    def write(self):
        self.config_cd.write()

    def read(self):
        fname = 'config.pkl'

        """ Read config file  """
        self.logger.debug(f'config read {self.cache_dir} {fname}')
        # Verify main directory exists
        if not os.path.exists(self.cache_dir):
            self.logger.warning(f"{self.cache_dir} folder not found.")
            Widge.fatal_error(f"{self.cache_dir} folder not found.  Please run GeoUtil.py to correct")

        self.config_cd = CachedDictionary.CachedDictionary(self.cache_dir, fname)
        self.config_cd.read()

        if self.config_cd.error:
            self.logger.error(f'Config {os.path.join(self.cache_dir, fname)} not found')

            # Create empty config file
            path = os.path.join(self.cache_dir, "config.pkl")
            self.set("gedcom_path", "GEDCOM filename: <empty>")
            with open(path, 'wb') as file:
                pickle.dump(self.config_cd.dict, file)
            return True
        else:
            return False

    def valid_directories(self)->bool:
        if not os.path.exists(self.cache_dir):
            return False
        else:
            return True

    def create_directories(self):
        if not os.path.exists(self.directory):
            self.logger.info(f'Creating main folder {self.directory}')
            os.makedirs(self.directory)

        if not os.path.exists(self.cache_dir):
            self.logger.info(f'Creating cache folder {self.cache_dir}')
            os.makedirs(self.cache_dir)





#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  Copyright (c) 2019.       Mike Herbert
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA

import logging
import os
import pickle
from pathlib import Path
from tkinter import *

from geofinder import CachedDictionary,  GeoKeys, UtilLayout

try:
    import unidecode
except ModuleNotFoundError:
    print('Unidecode missing.')


# Setup and modify config for geonames
class ReviewWindow:
    """
    Utilities to edit and configure items for Geofinder
    There are 5 tabs:
    
    Status - overall status
    Skiplist - for deleting items from the skiplist
    Replace - for deleting items from the replace list
    Files - for specifying which geo_name files to use for geoname data
    Countries - List of countries to include 
    
    Setup files/countries  allows users to:
     1. set country list config file.  Only entries for those countries will be read.    "country_list.pkl"
     2. select which geonames.org files to read.  file names are place in config file.   "file_names.pkl"
     3. If the above are changed, the utility deletes the cache file and the main GeoCoder app will rebuild it
    """

    def __init__(self):
        self.logger = logging.getLogger(__name__)
        fmt = "%(levelname)s %(asctime)s %(name)s.%(funcName)s %(lineno)d: %(message)s"
        logging.basicConfig(level=logging.INFO, format=fmt)
        self.logger.info('Setup')

        self.directory: str = os.path.join(str(Path.home()), GeoKeys.get_directory_name())
        self.cache_dir = GeoKeys.get_cache_directory()

        # Get configuration settings stored in config pickle file
        self.cfg: CachedDictionary.CachedDictionary = CachedDictionary.CachedDictionary(self.cache_dir, "config.pkl")

        if not os.path.exists(self.directory):
            self.logger.info(f'Creating main folder {self.directory}')
            os.makedirs(self.directory)

        if not os.path.exists(self.cache_dir):
            self.logger.info(f'Creating cache folder {self.cache_dir}')
            os.makedirs(self.cache_dir)

        self.cfg.read()

        # Verify config -  test to see if gedcom file accessible
        self.get_config()

        # Create App window
        self.root = Tk()
        self.root["padx"] = 30
        self.root["pady"] = 30
        self.root.title('GeoUtil')

        UtilLayout.UtilLayout(root=self.root, directory=self.directory, cache_dir=self.cache_dir)

    def get_config(self):
        """ Read config file  """
        # Verify main directory exists
        self.logger.debug('get config')
        if not os.path.exists(self.directory):
            self.logger.info(f'Geoname folder not found {self.directory}.  Creating')
            os.makedirs(self.directory)

        path = self.cache_dir
        if not os.path.exists(path):
            self.logger.info(f'Data folder not found {path}.  Creating')

            os.makedirs(path)

            # Create empty config file
            path = os.path.join(self.cache_dir, "config.pkl")
            self.logger.info(f'Creating config pickle file {path}.')

            self.cfg.set("gedcom_path", "No file selected")
            with open(path, 'wb') as file:
                pickle.dump(self.cfg.dict, file)

        path = self.cfg.get("gedcom_path")


def entry():
    ReviewWindow()


if __name__ == "__main__":
    entry()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  Copyright (c) 2019.       Mike Herbert
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA

from tkinter import *
from tkinter import ttk, messagebox
from typing import List, Dict


class Widge:
    """
    These  routines provide helper functions for Tkint including standardized get and set text.
    """

    @staticmethod
    def disable_buttons(button_list: List[ttk.Button]) -> None:
        for button in button_list:
            button.config(state="disabled")

    @staticmethod
    def enable_buttons(button_list: List[ttk.Button]) -> None:
        for button in button_list:
            button.config(state="normal")

    @staticmethod
    def fatal_error(msg: str) -> None:
        """ Fatal error -  Notify user and shutdown """
        messagebox.showerror("Error", msg)
        sys.exit()

    @staticmethod
    def exit_dialog(title, msg):
        if messagebox.askyesno(title, msg):
            sys.exit()

    # todo remove set_grid_postion from listboxframe
    import collections

    GridInfo = collections.namedtuple('GridInfo', 'col row xpad ypad sticky')
    GridDict = Dict[str, GridInfo]

    @staticmethod
    def set_grid_position(widget, name: str, grd: GridDict) -> None:
        """
        Lookup name of the widget in grd and position widget in a Tkinter grid layout using params in grd
        grd is a dictionary and
        each entry is a list of Col, row, xpad, ypad, sticky (optional).
        """
        # grd = {"title": [0, 0, 5, 10, "EW"], "progress": [0, 1, 5, 5, "EW"]}

        if grd[name][4] == " ":
            # No sticky value
            widget.grid(row=grd[name].row, column=grd[name][0], padx=grd[name][2],
                        pady=grd[name][3])
        else:
            widget.grid(row=grd[name][1], column=grd[name][0], padx=grd[name][2],
                        pady=grd[name][3], sticky=grd[name][4])


class CLabel(ttk.Label):
    """
    Have generic set_text and get_text
    """

    def __init__(self, parent, **kwargs):
        ttk.Label.__init__(self, parent, **kwargs)

    def get_text(self) -> str:
        """ Get the text of a widget """
        return self.cget("text")

    def set_text(self, text: str):
        """ Set the text of a widget """
        self.configure(text=text)


class CEntry(ttk.Entry):
    """
    Add support for Undo/Redo to TextEntry on Ctl-Z, shift Ctl-y.
    Have generic set_text and get_text
    """

    def __init__(self, parent, *args, **kwargs):
        ttk.Entry.__init__(self, parent, *args, **kwargs)
        self.changes = [""]
        self.steps = int()
        self.bind("<Control-z>", self.undo)
        self.bind("<Control-y>", self.redo)
        self.bind("<Key>", self.add_changes)

    def insert(self, idx, txt):
        self.changes = [""]
        self.steps = int()
        super().insert(idx, txt)

    def undo(self,  _):
        if self.steps != 0:
            self.steps -= 1
            self.delete(0, END)
            super().insert(END, self.changes[self.steps])

    def redo(self,  _):
        if self.steps < len(self.changes):
            self.delete(0, END)
            super().insert(END, self.changes[self.steps])
            self.steps += 1

    def add_changes(self,  _):
        if self.get() != self.changes[-1]:
            self.changes.append(self.get())
            self.steps += 1

    def set_text(self, text: str):
        """ Set the text of a widget """
        self.delete("0", END)
        self.insert(0, text)

    def get_text(self) -> str:
        """ Get the text of a widget """
        return self.get()


"""
    def add_to_layer(self, layer, command, coords, **kwargs):
        # Add support for Z-Layers.  Add item to specific layer 
        layer_tag = "layer %s" % layer
        if layer_tag not in self._layers: self._layers.append(layer_tag)
        tags = kwargs.setdefault("tags", [])
        tags.append(layer_tag)
        item_id = command(coords, **kwargs)
        self._adjust_layers()
        return item_id

    def _adjust_layers(self):
        # Adjust everything to appropriate layer 
        for layer in sorted(self._layers):
            self.canvas.lift(layer)
"""
